/*  This file is part of FediBlock-backend.
 *  Copyright © 2020 tastytea <tastytea@tastytea.de>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, version 3.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "files.hpp"
#include "fs-compat.hpp"

#include <cstdint>
#include <cstdlib>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>

namespace FediBlock::files
{

using std::getenv;
using std::ifstream;
using std::runtime_error;
using std::string;
using std::stringstream;
using std::uintmax_t;

string _tmpdir;

fs::path get_tmpdir()
{
    if (_tmpdir.empty())
    {
        _tmpdir = fs::temp_directory_path() / "fediblock-backend-XXXXXX";
        if (mkdtemp(&_tmpdir[0]) == nullptr) // mkdtemp() modifies _tmpdir.
        {
            throw runtime_error{"Could not create temporary directory: "
                                + _tmpdir};
        }
    }

    return _tmpdir;
}

void remove_tmpdir()
{
    if (fs::remove_all(get_tmpdir()) == static_cast<uintmax_t>(-1))
    {
        throw runtime_error{"Couldn't remove temporary directory: "
                            + get_tmpdir().string()};
    }
}

fs::path get_datadir()
{
    char *env{getenv("XDG_DATA_HOME")};
    if (env != nullptr)
    {
        const auto path{fs::path(env) / "fediblock-backend"};
        fs::create_directories(path);
        return path;
    }

    env = getenv("HOME");
    if (env != nullptr)
    {
        auto path{fs::path(env)};
#if defined(__linux__)
        path /= ".local/share/fediblock-backend";
#elif defined(__unix__)
        path /= ".fediblock-backend";
#endif
        fs::create_directories(path);
        return path;
    }

    throw runtime_error{"Could not find data dir"};
}

string get_access_token()
{
    ifstream file(get_datadir() / "gitea_access_token");
    if (!file.good())
    {
        throw runtime_error{"Could not read access token: "
                            + (get_datadir() / "gitea_access_token").string()};
    }
    stringstream ss;
    ss << file.rdbuf();

    return ss.str().substr(0, ss.str().find('\n'));
}

} // namespace FediBlock::files
